structure Specializer =
struct

open CommandA;
open ExpressionA;
open AbstractValue;

exception SpecializerError of string;

fun applyHd( AV N ) = AV N
  | applyHd( AV( C( v1, v2 ) ) ) = AV v1
  | applyHd( ACONS( av1, av2 ) ) = av1
  | applyHd( av1 ) = AHD( av1 )
;

fun applyTl( AV N ) = AV N
  | applyTl( AV( C( v1, v2 ) ) ) = AV v2
  | applyTl( ACONS( av1, av2 ) ) = av2
  | applyTl( av1 ) = ATL( av1 )
;

fun applyCons( AV v1, AV v2 ) = AV( C( v1, v2 ) )
  | applyCons( av1, av2 ) = ACONS( av1, av2 )
;

fun applyEq( AV v1, AV v2 ) =
    if v1 = v2
    then AV( C(N,N) )
    else AV( N )
  | applyEq( av1, av2 ) =
    AEQ( av1, av2 )
;

fun evaluate( VARA, var ) = 
    var
  | evaluate( VALA v, var ) = 
    AV v
  | evaluate( HDA( e ), var ) = 
    applyHd( evaluate( e, var ) )
  | evaluate( TLA( e ), var ) = 
    applyTl( evaluate( e, var ) )
  | evaluate( CONSA( e1, e2 ), var ) = 
    applyCons( evaluate( e1, var ), evaluate( e2, var ) )
  | evaluate( EQA( e1, e2 ), var ) = 
    applyEq( evaluate( e1, var ), evaluate( e2, var ) )
  | evaluate( LIFT( e ), var ) = 
    evaluate( e, var )
  | evaluate( STATIC( e ), var ) = 
    evaluate( e, var )
;

fun specializeExpression( VARA, var ) = 
    Expression.VAR
  | specializeExpression( VALA v, var ) = 
    Expression.VAL v
  | specializeExpression( HDA( e ), var ) = 
    Expression.HD( specializeExpression( e, var ) )
  | specializeExpression( TLA( e ), var ) = 
    Expression.TL( specializeExpression( e, var ) )
  | specializeExpression( CONSA( e1, e2 ), var ) =
    Expression.CONS( specializeExpression( e1, var ), 
                     specializeExpression( e2, var ) )
  | specializeExpression( EQA( e1, e2 ), var ) =
    Expression.EQ( specializeExpression( e1, var ), 
                   specializeExpression( e2, var ) )
  | specializeExpression( LIFT( e ), var ) = 
    let
        val av = evaluate( e, var )
    in
        Expression.VAL( AbstractValue.value( av ) )
    end
  | specializeExpression( STATIC( e ), var ) = 
    Expression.VAL N
;

fun specializeCommand( ASSIGN( e ), ls, code, 
                       unroll, acc, last_var ) =
    let
        val new_e = compact( substitute( acc, e ) );
        val new_var = evaluate( new_e, last_var );
        val new_e = specializeExpression( new_e, last_var );
    in
        ( Command.SEQ( code, Command.ASSIGN( new_e ) ), 
          unroll, VARA, new_var )
    end
  | specializeCommand( ASSIGNS( e ), ls, code, 
                       unroll, acc, last_var ) =
    let
        val new_acc = compact( substitute( acc, e ) );
    in
        ( code, unroll, new_acc, last_var )
    end
  | specializeCommand( SEQ( c1, c2 ), ls, code, 
                       unroll, acc, last_var ) =
    let
        val( code1, unroll1, acc1, last_var1 ) = 
           specializeCommand( c1, ls, code, unroll, acc, last_var );
    in
        specializeCommand( c2, ls, code1, unroll1, acc1, last_var1 )
    end

  | specializeCommand( LOOP( c ), ls, code, 
                       unroll, acc, last_var ) =
    let
        val( code1, unroll1, acc1, last_var1 ) = 
           specializeCommand( c, c :: ls, 
                              Command.ASSIGN( Expression.VAR ), 
                              unroll, acc, last_var );
    in
        ( Command.SEQ( code, Command.LOOP( code1 ) ), 
          unroll1, acc1, last_var1 )
    end

  | specializeCommand( IF2S( e, c1, c2 ), ls, code, 
                       unroll, acc, last_var ) =
    let
        val new_e = compact( substitute( acc, e ) );
        val av = evaluate( new_e, last_var );
    in
        if AbstractValue.isValue( av )
        then if av = AV N
             then specializeCommand( c2, ls, code, 
                                     unroll, acc, last_var )
             else specializeCommand( c1, ls, code, 
                                     unroll, acc, last_var )
        else raise SpecializerError( "IF2S condition[ " ^ 
                                     ExpressionA.toString( e ) ^ 
                                     " ] evaluating to [ " ^ 
                                     AbstractValue.toString( av ) ^ 
                                     " ] is not static\n" )
    end
  | specializeCommand( IF2D( e, c1, c2 ), ls, code, 
                       unroll, acc, last_var ) =
    let
        val new_e = compact( substitute( acc, e ) );
        val new_e = specializeExpression( new_e, last_var );
        val( code1, unroll1, acc1, last_var1 ) = 
           specializeCommand( c1, ls, 
                              Command.ASSIGN( Expression.VAR ), 
                              unroll, acc, last_var );
        val( code2, unroll2, acc2, last_var2 ) = 
           specializeCommand( c2, ls, 
                              Command.ASSIGN( Expression.VAR ), 
                              unroll1, acc, last_var );
    in
        ( Command.SEQ( code, Command.IF2( new_e, code1, code2 ) ), 
          unroll2, acc2, last_var2 )
    end

  | specializeCommand( REPEAT, ls, code, unroll, acc, last_var ) =
    ( Command.SEQ( code, Command.REPEAT ), unroll, acc, last_var )

  | specializeCommand( UNROLL, [], code, unroll, acc, last_var ) =
    raise SpecializerError( "UNROLL needs loop\n" )
  | specializeCommand( UNROLL, c :: ls, code, 
                       unroll, acc, last_var ) =
    let
        val new_var = evaluate( acc, last_var );
    in
        specializeCommand( c, c :: ls, code, unroll, acc, last_var )
    end

  | specializeCommand( ESCAPE, ls, code, 
                       unroll, acc, last_var ) =
    ( code, ( acc, last_var ) :: unroll, acc, last_var )

  | specializeCommand( UNROLL_LG, [], code, 
                       unroll, acc, last_var ) =
    raise SpecializerError( "UNROLL_LG needs loop\n" )
  | specializeCommand( UNROLL_LG, c :: ls, code, 
                       unroll, acc, last_var ) =
    let
        val( code1, unroll1, acc1, last_var1 ) =
           specializeCommand( c, c :: c :: ls, 
                              Command.ASSIGN( Expression.VAR ), 
                              unroll, acc, last_var );
    in
        if List.null( unroll1 )
        then ( Command.SEQ( code, Command.LOOP( code1 ) ), 
               [], acc1, last_var1 )
        else specializeCommand( c, c :: ls, 
                                Command.SEQ( 
                                code, 
                                Command.LOOP( code1 ) ), 
                                tl( unroll1 ), 
                                #1( hd( unroll1 ) ), 
                                #2( hd( unroll1 ) ) )
    end

fun specialize( prog, static ) =
    let
        val source = SEQ( ASSIGN( CONSA( VALA static, VARA ) ), 
                          prog );
        val source = Tools.fix( NormalizationA.normalize, source );

        val init_code = Command.ASSIGN( Expression.VAR );

        val target = #1( specializeCommand( source, 
                                            [], 
                                            init_code, 
                                            [], VARA, AD ) );


        val target = Tools.fix( Normalization.normalize, target );
        val target = 
            #1( Pruner.remove_unused( target, 
                                      [ ( Expression.VAR, 
                                          Expression.VAR ) ], 
                                      [], [] ) );

        val target = Tools.fix( Normalization.normalize, target );
        val target = Optimizer.optimize( target );
            
    in
        target
    end
    handle Interpreter.SyntaxException( msg ) =>
           raise SpecializerError( 
                 "Interpreter syntax exception: " ^ 
                 msg )
         | Interpreter.ValueException( msg ) =>
           raise SpecializerError( 
                 "Interpreter value exception: " ^ 
                 msg )
         | Optimizer.OptimizerError( msg ) =>
           raise SpecializerError( "OptimizerError: " ^ msg )
;

end
;
